@@ -4,6 +4,8 @@
     These functions are intended to be used for comparing encodes, filterchains,
     and other kinds of personal analysis.
 """
+from __future__ import annotations
+
 import math
 import random
 import warnings
@@ -11,10 +13,10 @@ from abc import ABC, abstractmethod
 from enum import IntEnum
 from itertools import groupby, zip_longest
 from typing import (Any, Callable, Dict, Iterable, Iterator, List, Literal,
-                    Optional, Sequence, Set, Tuple, TypeVar, Union, overload)
+                    Sequence, Set, Tuple, TypeVar, overload)
 
 import vapoursynth as vs
-import vsutil
+from vsutil import disallow_variable_format, disallow_variable_resolution, get_w, get_subsampling, depth
 
 from .dehardsub import hardsub_mask
 from .misc import get_matrix
@@ -49,7 +51,7 @@ class Comparer(ABC):
                             Determines where to place clip name using ``VideoNode.text.Text`` (Default: 7)
     """
     def __init__(self,
-                 clips: Union[Dict[str, vs.VideoNode], Sequence[vs.VideoNode]],
+                 clips: Dict[str, vs.VideoNode] | Sequence[vs.VideoNode],
                  /,
                  *,
                  label_alignment: int = 7
@@ -62,20 +64,20 @@ class Comparer(ABC):
             raise ValueError(f"Comparer: unexpected type {type(clips)} for `clips` argument")
 
         self.clips: List[vs.VideoNode] = list(clips.values()) if isinstance(clips, dict) else list(clips)
-        self.names: Optional[List[str]] = list(clips.keys()) if isinstance(clips, dict) else None
+        self.names: List[str] | None = list(clips.keys()) if isinstance(clips, dict) else None
 
         self.label_alignment = label_alignment
 
         self.num_clips = len(clips)
 
         widths: Set[int] = {clip.width for clip in self.clips}
-        self.width: Optional[int] = w if (w := widths.pop()) != 0 and len(widths) == 0 else None
+        self.width: int | None = w if (w := widths.pop()) != 0 and len(widths) == 0 else None
 
         heights: Set[int] = {clip.height for clip in self.clips}
-        self.height: Optional[int] = h if (h := heights.pop()) != 0 and len(heights) == 0 else None
+        self.height: int | None = h if (h := heights.pop()) != 0 and len(heights) == 0 else None
 
-        formats: Set[Optional[str]] = {getattr(clip.format, 'name', None) for clip in self.clips}
-        self.format: Optional[str] = formats.pop() if len(formats) == 1 else None
+        formats: Set[str | None] = {getattr(clip.format, 'name', None) for clip in self.clips}
+        self.format: str | None = formats.pop() if len(formats) == 1 else None
 
     def _marked_clips(self) -> List[vs.VideoNode]:
         """
@@ -121,7 +123,7 @@ class Stack(Comparer):
     """
 
     def __init__(self,
-                 clips: Union[Dict[str, vs.VideoNode], Sequence[vs.VideoNode]],
+                 clips: Dict[str, vs.VideoNode] | Sequence[vs.VideoNode],
                  /,
                  *,
                  direction: Direction = Direction.HORIZONTAL,
@@ -161,7 +163,7 @@ class Interleave(Comparer):
     """
 
     def __init__(self,
-                 clips: Union[Dict[str, vs.VideoNode], Sequence[vs.VideoNode]],
+                 clips: Dict[str, vs.VideoNode] | Sequence[vs.VideoNode],
                  /,
                  *,
                  label_alignment: int = 7
@@ -219,10 +221,10 @@ class Tile(Comparer):
     """
 
     def __init__(self,
-                 clips: Union[Dict[str, vs.VideoNode], Sequence[vs.VideoNode]],
+                 clips: Dict[str, vs.VideoNode] | Sequence[vs.VideoNode],
                  /,
                  *,
-                 arrangement: Optional[List[List[int]]] = None,
+                 arrangement: List[List[int]] | None = None,
                  label_alignment: int = 7
                  ) -> None:
         super().__init__(clips, label_alignment=label_alignment)
@@ -255,7 +257,7 @@ class Tile(Comparer):
         return core.std.StackVertical(rows)
 
     def _auto_arrangement(self) -> List[List[int]]:
-        def _grouper(iterable: Iterable[T], n: int, fillvalue: Optional[T] = None) -> Iterator[Tuple[T, ...]]:
+        def _grouper(iterable: Iterable[T], n: int, fillvalue: T | None = None) -> Iterator[Tuple[T, ...]]:
             args = [iter(iterable)] * n
             return zip_longest(*args, fillvalue=fillvalue)
 
@@ -286,7 +288,7 @@ class Split(Stack):
     """
 
     def __init__(self,
-                 clips: Union[Dict[str, vs.VideoNode], Sequence[vs.VideoNode]],
+                 clips: Dict[str, vs.VideoNode] | Sequence[vs.VideoNode],
                  /,
                  *,
                  direction: Direction = Direction.HORIZONTAL,
@@ -305,7 +307,7 @@ class Split(Stack):
                               or (self.direction == Direction.VERTICAL and (((self.height // self.num_clips) % 2)
                                                                             or ((self.height % self.num_clips) % 2))))
 
-        is_subsampled = not all(vsutil.get_subsampling(clip) in ('444', None) for clip in self.clips)
+        is_subsampled = not all(get_subsampling(clip) in ('444', None) for clip in self.clips)
 
         if breaks_subsampling and is_subsampled:
             raise ValueError("Split: resulting cropped width or height violates subsampling rules; "
@@ -337,8 +339,8 @@ class Split(Stack):
 
 
 def compare(clip_a: vs.VideoNode, clip_b: vs.VideoNode,
-            frames: Optional[List[int]] = None,
-            rand_total: Optional[int] = None,
+            frames: List[int] | None = None,
+            rand_total: int | None = None,
             force_resample: bool = True, print_frame: bool = True,
             mismatch: bool = False) -> vs.VideoNode:
     """
@@ -393,6 +395,8 @@ def compare(clip_a: vs.VideoNode, clip_b: vs.VideoNode,
     return core.std.Interleave([frames_a, frames_b], mismatch=mismatch)
 
 
+@disallow_variable_format
+@disallow_variable_resolution
 def stack_compare(*clips: vs.VideoNode,
                   make_diff: bool = True,
                   height: int = 288,
@@ -423,7 +427,7 @@ def stack_compare(*clips: vs.VideoNode,
         raise ValueError("stack_compare: `make_diff` only works for exactly 2 clips")
 
     clipa, clipb = clips
-    scaled_width = vsutil.get_w(height, only_even=False)
+    scaled_width = get_w(height, only_even=False)
 
     diff = core.std.MakeDiff(clipa=clipa, clipb=clipb)
     diff = diff.resize.Spline36(width=scaled_width * 2, height=height * 2).text.FrameNum(8)
@@ -433,6 +437,8 @@ def stack_compare(*clips: vs.VideoNode,
     return Stack([Stack(resized).clip, diff], direction=Direction.VERTICAL).clip
 
 
+@disallow_variable_format
+@disallow_variable_resolution
 def stack_planes(clip: vs.VideoNode, /, stack_vertical: bool = False) -> vs.VideoNode:
     """
     Stacks the planes of a clip.
@@ -445,12 +451,12 @@ def stack_planes(clip: vs.VideoNode, /, stack_vertical: bool = False) -> vs.Vide
 
     :return:               Clip with stacked planes
     """
-    if clip.format is None:
-        raise ValueError("stack_planes: variable-format clips not allowed")
+    assert clip.format
+
     if clip.format.num_planes != 3:
         raise ValueError("stack_planes: input clip must be in YUV or RGB planar format")
 
-    split_planes = vsutil.split(clip)
+    split_planes = split(clip)
 
     if clip.format.color_family == vs.ColorFamily.YUV:
         planes: Dict[str, vs.VideoNode] = {'Y': split_planes[0], 'U': split_planes[1], 'V': split_planes[2]}
@@ -461,10 +467,10 @@ def stack_planes(clip: vs.VideoNode, /, stack_vertical: bool = False) -> vs.Vide
 
     direction: Direction = Direction.HORIZONTAL if not stack_vertical else Direction.VERTICAL
 
-    if vsutil.get_subsampling(clip) in ('444', None):
+    if get_subsampling(clip) in ('444', None):
         return Stack(planes, direction=direction).clip
 
-    elif vsutil.get_subsampling(clip) == '420':
+    elif get_subsampling(clip) == '420':
         subsample_direction: Direction = Direction.HORIZONTAL if stack_vertical else Direction.VERTICAL
         y_plane = planes.pop('Y').text.Text(text='Y')
         subsampled_planes = Stack(planes, direction=subsample_direction).clip
@@ -472,9 +478,11 @@ def stack_planes(clip: vs.VideoNode, /, stack_vertical: bool = False) -> vs.Vide
         return Stack([y_plane, subsampled_planes], direction=direction).clip
 
     else:
-        raise ValueError(f"stack_planes: unexpected subsampling {vsutil.get_subsampling(clip)}")
+        raise ValueError(f"stack_planes: unexpected subsampling {get_subsampling(clip)}")
 
 
+@disallow_variable_format
+@disallow_variable_resolution
 def diff_hardsub_mask(a: vs.VideoNode, b: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
     """
     Diff func for :py:func:`lvsfunc.comparison.diff` to use a hardsub mask.
@@ -494,7 +502,7 @@ def diff(*clips: vs.VideoNode,
          height: int = ...,
          interleave: bool = ...,
          return_ranges: Literal[True] = True,
-         exclusion_ranges: Optional[Sequence[Union[int, Tuple[int, int]]]] = ...,
+         exclusion_ranges: Sequence[int | Tuple[int, int]] | None = ...,
          diff_func: Callable[[vs.VideoNode, vs.VideoNode], vs.VideoNode] = ...,
          **namedclips: vs.VideoNode) -> Tuple[vs.VideoNode, List[Tuple[int, int]]]:
     ...
@@ -506,20 +514,22 @@ def diff(*clips: vs.VideoNode,
          height: int = ...,
          interleave: bool = ...,
          return_ranges: Literal[False],
-         exclusion_ranges: Optional[Sequence[Union[int, Tuple[int, int]]]] = ...,
+         exclusion_ranges: Sequence[int | Tuple[int, int]] | None = ...,
          diff_func: Callable[[vs.VideoNode, vs.VideoNode], vs.VideoNode] = ...,
          **namedclips: vs.VideoNode) -> vs.VideoNode:
     ...
 
 
+@disallow_variable_format
+@disallow_variable_resolution
 def diff(*clips: vs.VideoNode,
          thr: float = 72,
          height: int = 288,
          interleave: bool = False,
          return_ranges: bool = False,
-         exclusion_ranges: Optional[Sequence[Union[int, Tuple[int, int]]]] = None,
+         exclusion_ranges: Sequence[int | Tuple[int, int]] | None = None,
          diff_func: Callable[[vs.VideoNode, vs.VideoNode], vs.VideoNode] = lambda a, b: core.std.MakeDiff(a, b),
-         **namedclips: vs.VideoNode) -> Union[vs.VideoNode, Tuple[vs.VideoNode, List[Tuple[int, int]]]]:
+         **namedclips: vs.VideoNode) -> vs.VideoNode | Tuple[vs.VideoNode, List[Tuple[int, int]]]:
     """
     Creates a standard :py:class:`lvsfunc.comparison.Stack` between frames from two clips that have differences.
     Useful for making comparisons between TV and BD encodes, as well as clean and hardsubbed sources.
@@ -578,10 +588,10 @@ def diff(*clips: vs.VideoNode,
             yield groupl[0][1], groupl[-1][1]
 
     if clips:
-        a, b = vsutil.depth(clips[0], 8), vsutil.depth(clips[1], 8)
+        a, b = depth(clips[0], 8), depth(clips[1], 8)
     elif namedclips:
         nc = list(namedclips.values())
-        a, b = vsutil.depth(nc[0], 8), vsutil.depth(nc[1], 8)
+        a, b = depth(nc[0], 8), depth(nc[1], 8)
 
     frames = []
     if thr <= 1:
@@ -628,7 +638,7 @@ def diff(*clips: vs.VideoNode,
         comparison = Interleave({f'{name_a}': core.std.Splice([a[f] for f in frames]),
                                  f'{name_b}': core.std.Splice([b[f] for f in frames])}).clip
     else:
-        scaled_width = vsutil.get_w(height, only_even=False)
+        scaled_width = get_w(height, only_even=False)
         diff = diff.resize.Spline36(width=scaled_width * 2, height=height * 2).text.FrameNum(9)
         a, b = (c.resize.Spline36(width=scaled_width, height=height).text.FrameNum(9) for c in (a, b))
 
@@ -660,6 +670,7 @@ def interleave(*clips: vs.VideoNode, **namedclips: vs.VideoNode) -> vs.VideoNode
     return Interleave(clips if clips else namedclips).clip
 
 
+@disallow_variable_resolution
 def split(*clips: vs.VideoNode, **namedclips: vs.VideoNode) -> vs.VideoNode:
     """
     Small convenience funciton for splitting clips along the x-axis and then stacking.
@@ -678,6 +689,7 @@ def split(*clips: vs.VideoNode, **namedclips: vs.VideoNode) -> vs.VideoNode:
     return Split(clips if clips else namedclips, label_alignment=2).clip
 
 
+@disallow_variable_resolution
 def stack_horizontal(*clips: vs.VideoNode, **namedclips: vs.VideoNode) -> vs.VideoNode:
     """
     Small convenience function for stacking clips horizontally.
@@ -693,6 +705,7 @@ def stack_horizontal(*clips: vs.VideoNode, **namedclips: vs.VideoNode) -> vs.Vid
     return Stack(clips if clips else namedclips).clip
 
 
+@disallow_variable_resolution
 def stack_vertical(*clips: vs.VideoNode, **namedclips: vs.VideoNode) -> vs.VideoNode:
     """
     Small convenience function for stacking clips vertically.
@@ -708,6 +721,7 @@ def stack_vertical(*clips: vs.VideoNode, **namedclips: vs.VideoNode) -> vs.Video
     return Stack(clips if clips else namedclips, direction=Direction.VERTICAL).clip
 
 
+@disallow_variable_resolution
 def tile(*clips: vs.VideoNode, **namedclips: vs.VideoNode) -> vs.VideoNode:
     """
     Small convenience function for tiling clips in a rectangular pattern.
