@@ -1,24 +1,29 @@
 """
     Deblocking functions and wrappers.
 """
+from __future__ import annotations
+
 from functools import partial
-from importlib.metadata import version
-from typing import Any, Optional, Sequence, Tuple, Union
+from typing import Any, Optional, Sequence, SupportsFloat, Tuple
 
 import vapoursynth as vs
-from packaging import version as pck_version
+from vsutil import (Dither, depth, disallow_variable_format,
+                    disallow_variable_resolution, get_depth)
 
+from .kernels import Catrom, Kernel
 from .types import Matrix
 from .util import get_prop
 
 core = vs.core
 
 
+@disallow_variable_format
+@disallow_variable_resolution
 def autodb_dpir(clip: vs.VideoNode, edgevalue: int = 24,
                 strs: Sequence[float] = [30, 50, 75],
                 thrs: Sequence[Tuple[float, float, float]] = [(1.5, 2.0, 2.0), (3.0, 4.5, 4.5), (5.5, 7.0, 7.0)],
-                matrix: Optional[Union[Matrix, int]] = None,
-                cuda: bool = True, device_index: int = 0,
+                matrix: Optional[Matrix | int] = None,
+                cuda: bool = True,
                 write_props: bool = False,
                 **vsdpir_args: Any) -> vs.VideoNode:
     """
@@ -45,6 +50,8 @@ def autodb_dpir(clip: vs.VideoNode, edgevalue: int = 24,
     :param edgevalue:       Remove edges from the edgemask that exceed this threshold (higher means more edges removed)
     :param strs:            A list of DPIR strength values (higher means stronger deblocking).
                             You can pass any arbitrary number of values here.
+                            Sane deblocking strenghts lie between 1–20 for most regular deblocking.
+                            Going higher than 50 is not recommended outside of very extreme cases.
                             The amount of values in strs and thrs need to be equal.
     :param thrs:            A list of thresholds, written as [(EdgeValRef, NextFrameDiff, PrevFrameDiff)].
                             You can pass any arbitrary number of values here.
@@ -52,15 +59,13 @@ def autodb_dpir(clip: vs.VideoNode, edgevalue: int = 24,
     :param matrix:          Enum for the matrix of the input clip. See ``types.Matrix`` for more info.
                             If `None`, gets matrix from the "_Matrix" prop of the clip unless it's an RGB clip,
                             in which case it stays as `None`.
-    :param cuda:            Device type used for deblocking. Uses CUDA if True, else CPU
-    :param device_index:    The 'device_index' + 1º device of type device type in the system
-    :write_props            Will write verbose props
-    :vsdpir_args            Additional args to pass to ``vsdpir``
+    :param cuda:            Use CUDA backend if True, else CPU backend
+    :param write_props:     Will write verbose props
+    :param vsdpir_args:     Additional args to pass to ``vsdpir``
 
     :return:                Deblocked clip
     """
-    if clip.format is None:
-        raise ValueError("autodb_dpir: 'Variable-format clips not supported'")
+    assert clip.format
 
     def _eval_db(n: int, f: Sequence[vs.VideoFrame],
                  clip: vs.VideoNode, db_clips: Sequence[vs.VideoNode],
@@ -117,7 +122,7 @@ def autodb_dpir(clip: vs.VideoNode, edgevalue: int = 24,
     diffprev = core.std.PlaneStats(rgb, rgb[0] + rgb, prop='YPrev')
 
     db_clips = [
-        vsdpir(rgb, strength=st, mode='deblock', cuda=cuda, device_index=device_index, **vsdpir_args)
+        vsdpir(rgb, strength=st, mode='deblock', cuda=cuda, **vsdpir_args)
         .std.SetFrameProp('Adb_DeblockStrength', intval=int(st)) for st in strs
     ]
 
@@ -129,60 +134,77 @@ def autodb_dpir(clip: vs.VideoNode, edgevalue: int = 24,
     return core.resize.Bicubic(debl, format=clip.format.id, matrix=matrix if not is_rgb else None)
 
 
-def vsdpir(clip: vs.VideoNode, strength: float = 25, mode: str = 'deblock',
-           matrix: Optional[Union[Matrix, int]] = None,
-           cuda: bool = True, device_index: int = 0,
-           i444: bool = False, **vsdpir_args: Any) -> vs.VideoNode:
+@disallow_variable_format
+@disallow_variable_resolution
+def vsdpir(clip: vs.VideoNode, strength: SupportsFloat | vs.VideoNode | None = 25, mode: str = 'deblock',
+           matrix: Matrix | int | None = None, tiles: int | Tuple[int] | None = None,
+           cuda: bool = True, i444: bool = False, kernel: Kernel = Catrom(), **dpir_args: Any) -> vs.VideoNode:
     """
-    A simple vs-dpir wrapper for convenience.
+    A simple vs-mlrt DPIR wrapper for convenience.
 
-    You must have vs-dpir 1.6.0 or higher.
-    https://github.com/HolyWu/vs-dpir.
+    You must install vs-mlrt. For more information, see the following links:
 
-    Converts to RGB -> runs vs-dpir -> converts back to original format.
+    * https://github.com/AmusementClub/vs-mlrt
+    * https://github.com/AmusementClub/vs-mlrt/wiki/DPIR
+    * https://github.com/AmusementClub/vs-mlrt/releases/latest
+
+    Converts to RGB -> runs DPIR -> converts back to original format, and with no subsampling if ``i444=True``.
     For more information, see https://github.com/cszn/DPIR.
 
     Dependencies:
 
-    * vs-dpir
+    * vs-mlrt
 
     :param clip:            Input clip
-    :param strength:        vs-dpir strength.
-                            Sane values lie between 20-50 for ``mode='deblock'``, and 2-5 for ``mode='denoise'``
-    :param mode:            vs-dpir mode. Valid modes are 'deblock' and 'denoise'.
+    :param strength:        DPIR strength. Sane values lie between 1–20 for ``mode='deblock'``,
+                            and 1–3 for ``mode='denoise'``
+    :param mode:            DPIR mode. Valid modes are 'deblock' and 'denoise'.
     :param matrix:          Enum for the matrix of the input clip. See ``types.Matrix`` for more info.
-                            If `None`, gets matrix from the "_Matrix" prop of the clip unless it's an RGB clip,
+                            If not specified, gets matrix from the "_Matrix" prop of the clip unless it's an RGB clip,
                             in which case it stays as `None`.
-    :param cuda:            Use CUDA if True, else CPU
+    :param cuda:            Use CUDA backend if True, else CPU backend
     :param i444:            Forces the returned clip to be YUV444PS instead of the input clip's format
-    :param vsdpir_args:     Additional args to pass onto vs-dpir
-                            (Note: strength, task, and device_type can't be overridden!)
+    :param dpir_args:       Additional args to pass to vs-mlrt.
+                            Note: strength, tiles, and model cannot be overridden!
 
     :return:                Deblocked or denoised clip in either the given clip's format or YUV444PS
     """
     try:
-        from vsdpir import DPIR
+        from vsmlrt import DPIR, Backend, DPIRModel
     except ModuleNotFoundError:
-        raise ModuleNotFoundError("'vsdpir: vs-dpir and pytorch are required to use this function.'")
+        raise ModuleNotFoundError("deblock: 'vsmlrt is required to use deblocking function.'")
 
-    # Forcing user to use a version of vsdpir that doesn't have the dotting issue rather than worry about that here.
-    if pck_version.parse(version('vsdpir')) < pck_version.parse("1.6.0"):
-        raise ImportError(f"vsdpir: 'Your vs-dpir version must be at least 1.6.0. Current version: {version('vsdpir')}"
-                          "\nPlease install the latest version from <https://github.com/HolyWu/vs-dpir>'")
+    assert clip.format
 
-    if clip.format is None:
-        raise ValueError("vsdpir: 'Variable-format clips not supported'")
+    bit_depth = get_depth(clip)
+    is_rgb, is_gray = (clip.format.color_family is f for f in (vs.RGB, vs.GRAY))
 
-    is_rgb = clip.format.color_family is vs.RGB
+    clip_32 = depth(clip, 32, dither_type=Dither.ERROR_DIFFUSION).std.Limiter()
 
-    if matrix is None and not is_rgb:
+    # TODO: Replace with a match-case?
+    if mode.lower() == 'deblock':
+        model = DPIRModel.drunet_deblocking_color if not is_gray else DPIRModel.drunet_deblocking_grayscale
+    elif mode.lower() == 'denoise':
+        model = DPIRModel.drunet_color if not is_gray else DPIRModel.drunet_gray
+    else:
+        raise ValueError(f"""vsdpir: '"{mode}" is not a valid mode!'""")
+
+    dpir_args |= dict(strength=strength, tiles=tiles, model=model)
+
+    if "backend" not in dpir_args:
+        dpir_args |= dict(backend=Backend.ORT_CUDA if cuda else Backend.OV_CPU)
+
+    if is_rgb or is_gray:
+        return depth(DPIR(clip_32, **dpir_args), bit_depth)
+
+    if matrix is None:
         matrix = get_prop(clip.get_frame(0), "_Matrix", int)
 
-    vsdpir_args |= dict(strength=strength, task=mode, device_type='cuda' if cuda else 'cpu')
+    targ_matrix = Matrix(matrix)
+    targ_format = clip.format.replace(subsampling_w=0, subsampling_h=0) if i444 else clip.format
 
-    clip_rgb = core.resize.Bicubic(clip, format=vs.RGBS, matrix_in=matrix, dither_type='error_diffusion')
-    run_dpir = DPIR(clip_rgb, **vsdpir_args)
+    clip_rgb = kernel.resample(clip_32, vs.RGBS, matrix_in=targ_matrix)  # type:ignore[arg-type]
 
-    if i444:
-        return core.resize.Bicubic(run_dpir, format=vs.YUV444PS, matrix=matrix)
-    return core.resize.Bicubic(run_dpir, format=clip.format.id, matrix=matrix)
+    run_dpir = DPIR(clip_rgb, **dpir_args)
+
+    return kernel.resample(run_dpir, targ_format, targ_matrix)  # type:ignore[arg-type]
