@@ -1,12 +1,14 @@
 """
     Dehardsubbing functions and wrappers.
 """
-import vapoursynth as vs
-
-import vsutil
+from __future__ import annotations
 
 from abc import ABC
-from typing import Any, List, Optional, Tuple, Union
+from typing import Any, List, Tuple
+
+import vapoursynth as vs
+from vsutil import (disallow_variable_format, disallow_variable_resolution,
+                    iterate, split)
 
 from .mask import DeferredMask
 from .types import Range
@@ -49,15 +51,15 @@ class HardsubMask(DeferredMask, ABC):
         thresh = scale_thresh(0.75, masks[-1])
         for p in partials:
             masks.append(core.std.Expr([masks[-1], self.get_mask(p, ref)], expr="x y -"))
-            dmasks.append(vsutil.iterate(core.std.Expr([masks[-1]], f"x {thresh} < 0 x ?"),
-                                         core.std.Maximum,
-                                         4).std.Inflate())
+            dmasks.append(iterate(core.std.Expr([masks[-1]], f"x {thresh} < 0 x ?"),
+                                  core.std.Maximum,
+                                  4).std.Inflate())
             pdhs.append(core.std.MaskedMerge(pdhs[-1], p, dmasks[-1]))
             masks[-1] = core.std.MaskedMerge(masks[-1], masks[-1].std.Invert(), masks[-2])
         return pdhs, dmasks
 
     def apply_dehardsub(self, hrdsb: vs.VideoNode, ref: vs.VideoNode,
-                        partials: Optional[List[vs.VideoNode]]) -> vs.VideoNode:
+                        partials: List[vs.VideoNode] | None) -> vs.VideoNode:
         """
         Apply dehardsubbing to a clip.
 
@@ -141,9 +143,9 @@ class HardsubLine(HardsubMask):
 
     :param expand: ``kgf.hardsubmask`` expand parameter (Default: clip.width // 200)
     """
-    expand: Optional[int]
+    expand: int | None
 
-    def __init__(self, *args: Any, expand: Optional[int] = None, **kwargs: Any) -> None:
+    def __init__(self, *args: Any, expand: int | None = None, **kwargs: Any) -> None:
         self.expand = expand
         try:
             from kagefunc import hardsubmask
@@ -167,7 +169,7 @@ class HardsubLineFade(HardsubLine):
     """
     ref_float: float
 
-    def __init__(self, ranges: Union[Range, List[Range]], *args: Any,
+    def __init__(self, ranges: Range | List[Range], *args: Any,
                  refframe: float = 0.5, **kwargs: Any) -> None:
         if refframe < 0 or refframe > 1:
             raise ValueError("HardsubLineFade: 'refframe must be between 0 and 1!'")
@@ -192,7 +194,7 @@ class HardsubSignFade(HardsubSign):
     """
     ref_float: float
 
-    def __init__(self, ranges: Union[Range, List[Range]], *args: Any,
+    def __init__(self, ranges: Range | List[Range], *args: Any,
                  refframe: float = 0.5, **kwargs: Any) -> None:
         if refframe < 0 or refframe > 1:
             raise ValueError("HardsubSignFade: 'refframe must be between 0 and 1!'")
@@ -215,11 +217,11 @@ class HardsubASS(HardsubMask):
                      May misbehave due to timestamp rounding.
     """
     filename: str
-    fontdir: Optional[str]
-    shift: Optional[int]
+    fontdir: str | None
+    shift: int | None
 
-    def __init__(self, filename: str, *args: Any, fontdir: Optional[str] = None,
-                 shift: Optional[int] = None, **kwargs: Any) -> None:
+    def __init__(self, filename: str, *args: Any, fontdir: str | None = None,
+                 shift: int | None = None, **kwargs: Any) -> None:
         self.filename = filename
         self.fontdir = fontdir
         self.shift = shift
@@ -230,8 +232,8 @@ class HardsubASS(HardsubMask):
         mask = ref.sub.TextFile(self.filename, fontdir=self.fontdir, blend=False)[1]  # horrific abuse of typechecker
         mask = mask[self.shift:] if self.shift else mask
         mask = mask.std.Binarize(1)
-        mask = vsutil.iterate(mask, core.std.Maximum, 3)
-        mask = vsutil.iterate(mask, core.std.Inflate, 3)
+        mask = iterate(mask, core.std.Maximum, 3)
+        mask = iterate(mask, core.std.Inflate, 3)
         return mask
 
 
@@ -255,7 +257,7 @@ def get_all_masks(hrdsb: vs.VideoNode, ref: vs.VideoNode, signs: List[HardsubMas
 
 
 def bounded_dehardsub(hrdsb: vs.VideoNode, ref: vs.VideoNode, signs: List[HardsubMask],
-                      partials: Optional[List[vs.VideoNode]] = None) -> vs.VideoNode:
+                      partials: List[vs.VideoNode] | None = None) -> vs.VideoNode:
     """
     Apply a list of :py:class:`lvsfunc.dehardsub.HardsubSign`
 
@@ -272,6 +274,8 @@ def bounded_dehardsub(hrdsb: vs.VideoNode, ref: vs.VideoNode, signs: List[Hardsu
     return hrdsb
 
 
+@disallow_variable_format
+@disallow_variable_resolution
 def hardsub_mask(hrdsb: vs.VideoNode, ref: vs.VideoNode, thresh: float = 0.06,
                  minimum: int = 1, expand: int = 8, inflate: int = 7) -> vs.VideoNode:
     """
@@ -286,15 +290,14 @@ def hardsub_mask(hrdsb: vs.VideoNode, ref: vs.VideoNode, thresh: float = 0.06,
 
     :return:        Hardsub mask
     """
-    if hrdsb.format is None:
-        raise ValueError("hardsub_mask: 'Variable-format clips not supported'")
+    assert hrdsb.format
 
     hsmf = core.std.Expr([hrdsb, ref], 'x y - abs') \
         .resize.Point(format=hrdsb.format.replace(subsampling_w=0, subsampling_h=0).id)
-    hsmf = core.std.Expr(vsutil.split(hsmf), "x y z max max")
+    hsmf = core.std.Expr(split(hsmf), "x y z max max")
     hsmf = hsmf.std.Binarize(scale_thresh(thresh, hsmf))
-    hsmf = vsutil.iterate(hsmf, core.std.Minimum, minimum)
-    hsmf = vsutil.iterate(hsmf, core.std.Maximum, expand)
-    hsmf = vsutil.iterate(hsmf, core.std.Inflate, inflate)
+    hsmf = iterate(hsmf, core.std.Minimum, minimum)
+    hsmf = iterate(hsmf, core.std.Maximum, expand)
+    hsmf = iterate(hsmf, core.std.Inflate, inflate)
 
     return hsmf
