@@ -8,9 +8,10 @@
     Don't tell louis this exists though, else I'm a dead man.
 """
 
-from typing import Optional
+from __future__ import annotations
 
 import vapoursynth as vs
+from vsutil import disallow_variable_resolution
 
 from .kernels import Point
 from .util import force_mod
@@ -18,7 +19,8 @@ from .util import force_mod
 core = vs.core
 
 
-def minecraftify(clip: vs.VideoNode, div: float = 64.0, mod: Optional[int] = None) -> vs.VideoNode:
+@disallow_variable_resolution
+def minecraftify(clip: vs.VideoNode, div: float = 64.0, mod: int | None = None) -> vs.VideoNode:
     """
     Function that transforms your clip into a Minecraft.
     Idea from Meme-Maji's Kobayashi memery (love you varde).
