@@ -1,20 +1,24 @@
 """
     Miscellaneous functions and wrappers that don't really have a place elsewhere.
 """
+from __future__ import annotations
+
 import colorsys
 import os
 import random
 import warnings
 from functools import partial, wraps
-from typing import (Any, Callable, Dict, List, Optional, Sequence, Tuple,
-                    TypeVar, Union, cast)
+from typing import Any, Callable, Dict, List, Sequence, Tuple, TypeVar, cast
 
 import vapoursynth as vs
-from vsutil import get_depth, get_w, get_y, is_image, scale_value
+from vsutil import (depth, disallow_variable_format,
+                    disallow_variable_resolution, get_depth, get_w, get_y,
+                    is_image, scale_value)
 
+from .kernels import Catrom
 from .mask import BoundingBox
-from .types import Position, Size
-from .util import get_prop
+from .types import Matrix, Position, Range, Size
+from .util import get_prop, normalize_ranges
 from .util import replace_ranges as _replace_ranges
 from .util import scale_thresh as _scale_thresh
 
@@ -25,7 +29,7 @@ core = vs.core
 annoying_formats = ['.iso', '.ts', '.vob']
 
 
-def source(file: str, ref: Optional[vs.VideoNode] = None,
+def source(file: str, ref: vs.VideoNode | None = None,
            force_lsmas: bool = False,
            mpls: bool = False, mpls_playlist: int = 0, mpls_angle: int = 0,
            **index_args: Any) -> vs.VideoNode:
@@ -95,7 +99,8 @@ def source(file: str, ref: Optional[vs.VideoNode] = None,
 
     if ref:
         if ref.format is None:
-            raise ValueError("source: 'Variable-format clips not supported.'")
+            raise ValueError("source: 'Variable-format ref clips not supported.'")
+
         clip = core.std.AssumeFPS(clip, fpsnum=ref.fps.numerator, fpsden=ref.fps.denominator)
         clip = core.resize.Bicubic(clip, width=ref.width, height=ref.height,
                                    format=ref.format.id, matrix=get_matrix(ref))
@@ -105,12 +110,14 @@ def source(file: str, ref: Optional[vs.VideoNode] = None,
     return clip
 
 
+@disallow_variable_format
+@disallow_variable_resolution
 def edgefixer(clip: vs.VideoNode,
-              left: Union[int, List[int], None] = None,
-              right: Union[int, List[int], None] = None,
-              top: Union[int, List[int], None] = None,
-              bottom: Union[int, List[int], None] = None,
-              radius: Optional[List[int]] = None,
+              left: int | List[int] | None = None,
+              right: int | List[int] | None = None,
+              top: int | List[int] | None = None,
+              bottom: int | List[int] | None = None,
+              radius: List[int] | None = None,
               full_range: bool = False) -> vs.VideoNode:
     """
     A wrapper for ContinuityFixer (https://github.com/MonoS/VS-ContinuityFixer).
@@ -155,6 +162,8 @@ def edgefixer(clip: vs.VideoNode,
     return ef if full_range else core.std.Limiter(ef, 16.0, [235, 240])
 
 
+@disallow_variable_format
+@disallow_variable_resolution
 def get_matrix(clip: vs.VideoNode) -> int:
     """
     Helper function to get the matrix for a clip.
@@ -163,6 +172,8 @@ def get_matrix(clip: vs.VideoNode) -> int:
 
     :return:        Value representing a matrix
     """
+    assert clip.format
+
     frame = clip.get_frame(0)
     w, h = frame.width, frame.height
 
@@ -175,7 +186,9 @@ def get_matrix(clip: vs.VideoNode) -> int:
     return 9
 
 
-def shift_tint(clip: vs.VideoNode, values: Union[int, Sequence[int]] = 16) -> vs.VideoNode:
+@disallow_variable_format
+@disallow_variable_resolution
+def shift_tint(clip: vs.VideoNode, values: int | Sequence[int] = 16) -> vs.VideoNode:
     """
     A function for forcibly adding pixel values to a clip.
     Can be used to fix green tints in Crunchyroll sources, for example.
@@ -215,7 +228,7 @@ def shift_tint(clip: vs.VideoNode, values: Union[int, Sequence[int]] = 16) -> vs
 
 
 def limit_dark(clip: vs.VideoNode, filtered: vs.VideoNode,
-               threshold: float = 0.25, threshold_range: Optional[int] = None) -> vs.VideoNode:
+               threshold: float = 0.25, threshold_range: int | None = None) -> vs.VideoNode:
     """
     Replaces frames in a clip with a filtered clip when the frame's darkness exceeds the threshold.
     This way you can run lighter (or heavier) filtering on scenes that are almost entirely dark.
@@ -232,7 +245,7 @@ def limit_dark(clip: vs.VideoNode, filtered: vs.VideoNode,
     """
     def _diff(n: int, f: vs.VideoFrame, clip: vs.VideoNode,
               filtered: vs.VideoNode, threshold: float,
-              threshold_range: Optional[int]) -> vs.VideoNode:
+              threshold_range: int | None) -> vs.VideoNode:
         psa = get_prop(f, "PlaneStatsAverage", float)
         if threshold_range:
             return filtered if threshold_range <= psa <= threshold else clip
@@ -247,10 +260,12 @@ def limit_dark(clip: vs.VideoNode, filtered: vs.VideoNode,
                                             threshold=threshold, threshold_range=threshold_range), avg)
 
 
+@disallow_variable_format
+@disallow_variable_resolution
 def wipe_row(clip: vs.VideoNode,
-             ref: Optional[vs.VideoNode] = None,
-             pos: Union[Position, Tuple[int, int]] = (1, 1),
-             size: Union[Size, Tuple[int, int], None] = None,
+             ref: vs.VideoNode | None = None,
+             pos: Position | Tuple[int, int] = (1, 1),
+             size: Size | Tuple[int, int] | None = None,
              show_mask: bool = False
              ) -> vs.VideoNode:
     """
@@ -324,8 +339,8 @@ def frames_since_bookmark(clip: vs.VideoNode, bookmarks: List[int]) -> vs.VideoN
 F = TypeVar("F", bound=Callable[..., vs.VideoNode])
 
 
-def allow_variable(width: Optional[int] = None, height: Optional[int] = None,
-                   format: Optional[int] = None
+def allow_variable(width: int | None = None, height: int | None = None,
+                   format: int | None = None
                    ) -> Callable[[Callable[..., vs.VideoNode]], Callable[..., vs.VideoNode]]:
     """
     Decorator allowing a variable-res and/or variable-format clip to be passed
@@ -396,7 +411,7 @@ def chroma_injector(func: F) -> F:
                                          colorfamily=vs.YUV)
             return res
 
-        out_fmt: Optional[vs.Format] = None
+        out_fmt: vs.VideoFormat | None = None
         if clip.format is not None:
             if clip.format.color_family not in (vs.GRAY, vs.YUV):
                 raise ValueError("chroma_injector: only YUV and GRAY clips are supported")
@@ -439,7 +454,7 @@ def chroma_injector(func: F) -> F:
 def colored_clips(amount: int,
                   max_hue: int = 300,
                   rand: bool = True,
-                  seed: Union[bytearray, bytes, float, int, str, None] = None,
+                  seed: bytearray | bytes | float | str | None = None,
                   **kwargs: Any
                   ) -> List[vs.VideoNode]:
     """
@@ -472,7 +487,7 @@ def colored_clips(amount: int,
 
     blank_clip_args: Dict[str, Any] = {'keep': 1, **kwargs}
 
-    hues: List[Union[float, int]] = [i * max_hue / (amount - 1) for i in range(amount - 1)]
+    hues: List[float] = [i * max_hue / (amount - 1) for i in range(amount - 1)]
     hues.append(max_hue)
 
     hls_color_list: List[Tuple[float, float, float]] = [colorsys.hls_to_rgb(h / 360, 0.5, 1) for h in hues]
@@ -485,6 +500,137 @@ def colored_clips(amount: int,
     return [core.std.BlankClip(color=color, **blank_clip_args) for color in rgb_color_list]
 
 
+@disallow_variable_format
+@disallow_variable_resolution
+def unsharpen(clip: vs.VideoNode, strength: float = 1.0, sigma: float = 1.5,
+              prefilter: bool = True, prefilter_sigma: float = 0.75) -> vs.VideoNode:
+    """
+    Diff'd unsharpening function.
+    Performs one-dimensional sharpening as such: "Original + (Original - blurred) * Strength"
+    It then merges back noise and detail that was prefiltered away,
+
+    This function also tries to mask haloing created by the unsharpening.
+    You can disable this by setting `mask` to False.
+
+    This function is not recommended for normal use,
+    but may be useful as prefiltering for detail- or edgemasks.
+
+    :param clip:                Input clip.
+    :param strength:            Amount to multiply blurred clip with original clip by
+    :param sigma:               Sigma for the gaussian blur.
+    :param prefilter:           Pre-denoising to prevent the unsharpening from picking up random noise.
+    :param prefilter_sigma:     Strength for the pre-denoising.
+    :param mask:                Use a halo mask to reduce haloing.
+    :param show_mask:           Show halo mask.
+
+    :return:                    Unsharpened clip
+    """
+    den = clip.dfttest.DFTTest(sigma=prefilter_sigma) if prefilter else clip
+    diff = core.std.MakeDiff(clip, den)
+
+    blurred_clip = core.bilateral.Gaussian(den, sigma=sigma)
+    unsharp = core.std.Expr(clips=[den, blurred_clip], expr=[f'x y - {strength} * x +', "", ""])
+    return core.std.MergeDiff(unsharp, diff)
+
+
+@disallow_variable_format
+@disallow_variable_resolution
+def overlay_sign(clip: vs.VideoNode, overlay: vs.VideoNode | str,
+                 frame_ranges: Range | List[Range] | None = None, fade_length: int = 0,
+                 matrix: Matrix | int | None = None) -> vs.VideoNode:
+    """
+    Wrapper to overlay a logo or sign onto another clip. Rewrite of fvsfunc.InsertSign.
+    This wrapper also allows you to set fades to fade a logo in and out.
+
+    Requires:
+
+    * vs-imwri
+
+    :param clip:            Base clip
+    :param overlay:         Sign or logo to overlay. Must be the png loaded in through imwri.Read()
+                            or a path string to the image file.
+    :param frame_ranges:    Frame ranges or starting frame to apply the overlay to. See ``types.Range`` for more info.
+                            If None, overlays the entire clip.
+                            If a Range is passed, the overlaid clip will only show up inside that range.
+                            If only a single integer is given, it will start on that frame and
+                            stay until the end of the clip.
+                            Note that this function only accepts a single Range! You can't pass a list of them!
+    :param fade_length:     Length to fade the clips into each other.
+                            The fade will start and end on the frames given in frame_ranges.
+                            If set to 0, it won't fade and the sign will simply pop in.
+    :param matrix:          Enum for the matrix of the input clip. See ``types.Matrix`` for more info.
+                            If not specified, gets matrix from the "_Matrix" prop of the clip unless it's an RGB clip,
+                            in which case it stays as `None`.
+
+    :return:                Clip with a logo or sign overlaid ontop for the given frame ranges,
+                            either with or without a fade
+    """
+    try:
+        from kagefunc import crossfade
+    except ModuleNotFoundError:
+        raise ModuleNotFoundError("overlay_sign: 'missing dependency `kagefunc`'")
+
+    assert clip.format
+
+    ov_type = type(overlay)
+    clip_fam = clip.format.color_family
+
+    # TODO: This can probably be done better
+    if not isinstance(overlay, (vs.VideoNode, str)):
+        raise ValueError("overlay_sign: '`overlay` must be a VideoNode object or a string path!'")
+    elif isinstance(overlay, str):
+        overlay = core.imwri.Read(overlay, alpha=True)
+
+    if not all([clip.width, overlay.width]) or not all([clip.height, overlay.height]):
+        raise ValueError("overlay_sign: 'Your overlay clip must have the same dimensions as your input clip!'")
+
+    # wtf mypy this IS reachable
+    if isinstance(frame_ranges, list) and len(frame_ranges) > 1:
+        warnings.warn("overlay_sign: 'Only one range is currently supported! "
+                      "Grabbing the first item in list.'")
+        frame_ranges = frame_ranges[0]
+
+    overlay = overlay[0] * clip.num_frames
+
+    if matrix is None:
+        matrix = get_prop(clip.get_frame(0), "_Matrix", int)
+
+    if matrix == 2:
+        raise ValueError("overlay_sign: 'You can't set a matrix of 2! "
+                         "Please set the correct matrix in the parameters!'")
+
+    assert overlay.format
+
+    if overlay.format.color_family is not clip_fam:
+        if clip_fam is vs.RGB:
+            overlay = Catrom().resample(overlay, clip.format.id, matrix_in=matrix)  # type:ignore[arg-type]
+        else:
+            overlay = Catrom().resample(overlay, clip.format.id, matrix)  # type:ignore[arg-type]
+
+    try:
+        mask = core.std.PropToClip(overlay)
+    except vs.Error:
+        if ov_type is str:
+            raise ValueError("overlay_sign: 'Please make sure your image has an alpha channel!'")
+        else:
+            raise ValueError("overlay_sign: 'Please make sure you loaded your sign in using imwri.Read!'")
+
+    merge = core.std.MaskedMerge(clip, overlay, depth(mask, get_depth(overlay)))
+
+    if not frame_ranges:
+        return merge
+
+    if fade_length > 0:
+        if isinstance(frame_ranges, int):
+            return crossfade(clip[:frame_ranges+fade_length], merge[frame_ranges:], fade_length)
+        else:
+            start, end = normalize_ranges(clip, frame_ranges)[0]
+            merge = crossfade(clip[:start+fade_length], merge[start:], fade_length)
+            return crossfade(merge[:end], clip[end-fade_length:], fade_length)
+    else:
+        return replace_ranges(clip, merge, frame_ranges)
+
+
 # compatibility since these were moved to util
 replace_ranges = _replace_ranges
 scale_thresh = _scale_thresh
