@@ -1,14 +1,21 @@
 """
-    (De)scaling functions and wrappers.
+    (De)scaling and conversion functions and wrappers.
 """
+from __future__ import annotations
+
 import math
 from functools import partial
-from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple, Union, cast
+from typing import Any, Callable, Dict, List, Literal, NamedTuple, cast
 
 import vapoursynth as vs
-from vsutil import depth, get_depth, get_w, get_y, iterate, join, plane
+from vsutil import (depth, disallow_variable_format,
+                    disallow_variable_resolution, get_depth, get_w, get_y,
+                    iterate, join, plane)
 
-from . import kernels, util
+from . import kernels
+from .kernels import Bicubic, BicubicSharp, Catrom, Kernel, Spline36
+from .types import VSFunction
+from .util import get_coefs, get_prop, quick_resample
 
 try:
     from cytoolz import functoolz
@@ -64,7 +71,7 @@ def _perform_descale(resolution: Resolution, clip: vs.VideoNode,
     return ScaleAttempt(descaled, rescaled, resolution, diff)
 
 
-def _select_descale(n: int, f: Union[vs.VideoFrame, List[vs.VideoFrame]],
+def _select_descale(n: int, f: vs.VideoFrame | List[vs.VideoFrame],
                     threshold: float, clip: vs.VideoNode,
                     clips_by_resolution: Dict[int, ScaleAttempt]
                     ) -> vs.VideoNode:
@@ -72,21 +79,21 @@ def _select_descale(n: int, f: Union[vs.VideoFrame, List[vs.VideoFrame]],
         f = [f]
     f = cast(List[vs.VideoFrame], f)
     best_res = max(f, key=lambda frame:
-                   math.log(clip.height - util.get_prop(frame, "descaleResolution", int), 2)
-                   * round(1 / max(util.get_prop(frame, "PlaneStatsAverage", float), 1e-12))
+                   math.log(clip.height - get_prop(frame, "descaleResolution", int), 2)
+                   * round(1 / max(get_prop(frame, "PlaneStatsAverage", float), 1e-12))
                    ** 0.2)
 
-    best_attempt = clips_by_resolution[util.get_prop(best_res, "descaleResolution", int)]
+    best_attempt = clips_by_resolution[get_prop(best_res, "descaleResolution", int)]
     if threshold == 0:
         return best_attempt.descaled
-    if util.get_prop(best_res, "PlaneStatsAverage", float) > threshold:
+    if get_prop(best_res, "PlaneStatsAverage", float) > threshold:
         return clip
     return best_attempt.descaled
 
 
 @functoolz.curry
 def reupscale(clip: vs.VideoNode,
-              width: Optional[int] = None, height: int = 1080,
+              width: int | None = None, height: int = 1080,
               kernel: kernels.Kernel = kernels.Bicubic(b=0, c=1/2),
               **kwargs: Any) -> vs.VideoNode:
     """
@@ -109,26 +116,25 @@ def reupscale(clip: vs.VideoNode,
 
     :return:             Reupscaled clip
     """
-
     width = width or get_w(height)
 
     # Doubling and downscale to given "h"
     znargs = dict(nsize=4, nns=4, qual=2, pscrn=2)
     znargs.update(kwargs)
 
-    upsc = util.quick_resample(clip, partial(core.znedi3.nnedi3, field=0,
-                                             dh=True, **znargs))
+    upsc = quick_resample(clip, partial(core.znedi3.nnedi3, field=0, dh=True, **znargs))
     upsc = core.std.FrameEval(upsc, partial(_transpose_shift, clip=upsc,
                                             kernel=kernel,
                                             caller="reupscale"),
                               prop_src=upsc)
-    upsc = util.quick_resample(upsc, partial(core.znedi3.nnedi3, field=0,
-                                             dh=True, **znargs))
+    upsc = quick_resample(upsc, partial(core.znedi3.nnedi3, field=0, dh=True, **znargs))
     return kernel.scale(upsc, width=height, height=width, shift=(0.5, 0)) \
         .std.Transpose()
 
 
 @functoolz.curry
+@disallow_variable_format
+@disallow_variable_resolution
 def descale_detail_mask(clip: vs.VideoNode, rescaled_clip: vs.VideoNode,
                         threshold: float = 0.05) -> vs.VideoNode:
     """
@@ -149,15 +155,16 @@ def descale_detail_mask(clip: vs.VideoNode, rescaled_clip: vs.VideoNode,
     return iterate(mask, core.std.Inflate, 2)
 
 
+@disallow_variable_format
+@disallow_variable_resolution
 def descale(clip: vs.VideoNode,
             upscaler:
-            Optional[Callable[[vs.VideoNode, int, int], vs.VideoNode]]
-            = reupscale,
-            width: Union[int, List[int], None] = None,
-            height: Union[int, List[int]] = 720,
+            Callable[[vs.VideoNode, int, int], vs.VideoNode] | None = reupscale,
+            width: int | List[int] | None = None,
+            height: int | List[int] = 720,
             kernel: kernels.Kernel = kernels.Bicubic(b=0, c=1/2),
             threshold: float = 0.0,
-            mask: Optional[Callable[[vs.VideoNode, vs.VideoNode], vs.VideoNode]]
+            mask: Callable[[vs.VideoNode, vs.VideoNode], vs.VideoNode] | None
             = descale_detail_mask, src_left: float = 0.0, src_top: float = 0.0,
             show_mask: bool = False) -> vs.VideoNode:
     """
@@ -203,8 +210,7 @@ def descale(clip: vs.VideoNode,
 
     :return:                       Descaled and re-upscaled clip with float bitdepth
     """
-    if clip.format is None:
-        raise ValueError("descale: 'Variable-format clips not supported'")
+    assert clip.format
 
     if type(height) is int:
         height = [height]
@@ -288,53 +294,224 @@ def descale(clip: vs.VideoNode,
     return join([upscaled, plane(clip, 1), plane(clip, 2)])
 
 
-def test_descale(clip: vs.VideoNode,
-                 width: Optional[int] = None, height: int = 720,
-                 kernel: kernels.Kernel = kernels.Bicubic(b=0, c=1/2),
-                 show_error: bool = True) -> Tuple[vs.VideoNode, ScaleAttempt]:
+CURVES = Literal[
+    vs.TransferCharacteristics.TRANSFER_IEC_61966_2_1,
+    vs.TransferCharacteristics.TRANSFER_BT709,
+    vs.TransferCharacteristics.TRANSFER_BT601,
+    vs.TransferCharacteristics.TRANSFER_ST240_M,
+    vs.TransferCharacteristics.TRANSFER_BT2020_10,
+    vs.TransferCharacteristics.TRANSFER_BT2020_12,
+]
+
+
+@disallow_variable_format
+@disallow_variable_resolution
+def ssim_downsample(clip: vs.VideoNode, width: int | None = None, height: int = 720,
+                    smooth: float | VSFunction = ((3 ** 2 - 1) / 12) ** 0.5,
+                    kernel: Kernel = Catrom(), gamma: bool = False,
+                    curve: CURVES = vs.TransferCharacteristics.TRANSFER_BT709,
+                    sigmoid: bool = False, epsilon: float = 1e-6) -> vs.VideoNode:
     """
-    Generic function to test descales with;
-    descales and reupscales a given clip, allowing you to compare the two easily.
-    Also returns a :py:class:`lvsfunc.scale.ScaleAttempt` with additional
-    information.
+    muvsfunc.ssim_downsample rewrite taken from a Vardë gist.
+    Unlike muvsfunc's implementation, this function also works in float and does not use nnedi3_resample.
+    Most of the documentation is taken from muvsfunc.
 
-    When comparing, it is recommended to do atleast a 4x zoom using Nearest Neighbor.
-    I also suggest using 'compare' (:py:func:`lvsfunc.comparison.compare`),
-    as that will make comparing the output with the source clip a lot easier.
+    SSIM downsampler is an image downscaling technique that aims to optimize
+    for the perceptual quality of the downscaled results.
+    Image downscaling is considered as an optimization problem
+    where the difference between the input and output images is measured
+    using famous Structural SIMilarity (SSIM) index.
+    The solution is derived in closed-form, which leads to the simple, efficient implementation.
+    The downscaled images retain perceptually important features and details,
+    resulting in an accurate and spatio-temporally consistent representation of the high resolution input.
 
-    Some of this code was leveraged from DescaleAA found in fvsfunc.
+    Original gist: https://gist.github.com/Ichunjo/16ab1f893588aafcb096c1f35a0cfb15
 
-    Dependencies:
+    :param clip:        Input clip
+    :param width:       Output width. If None, autocalculates using height
+    :param height:      Output height (default: 720)
+    :param smooth:      Image smoothening method.
+                        If you pass an int, it specifies the "radius" of the internel used boxfilter,
+                        i.e. the window has a size of (2*smooth+1)x(2*smooth+1).
+                        If you pass a float, it specifies the "sigma" of core.tcanny.TCanny,
+                        i.e. the standard deviation of gaussian blur.
+                        If you pass a function, it acts as a general smoother.
+                        Default uses a gaussian blur.
+    :param curve:       Gamma mapping
+    :param sigmoid:     When True, applies a sigmoidal curve after the power-like curve
+                        (or before when converting from linear to gamma-corrected).
+                        This helps reduce the dark halo artefacts found around sharp edges
+                        caused by resizing in linear luminance.
+    :param epsilon:     Machine epsilon
 
-    * vapoursynth-descale
-
-    :param clip:           Input clip
-    :param width:          Target descale width. If None, determine from `height`
-    :param height:         Target descale height (Default: 720)
-    :param kernel:         Kernel used to descale (see :py:class:`lvsfunc.kernels.Kernel`,
-                           Default: kernels.Bicubic(b=0, c=1/2))
-    :param show_error:     Render PlaneStatsDiff on the reupscaled frame (Default: True)
-
-    :return: A tuple containing a clip re-upscaled with the same kernel and
-             a ScaleAttempt tuple.
+    :return: Downsampled clip
     """
-    if clip.format is None:
-        raise ValueError("test_descale: 'Variable-format clips not supported'")
-
-    width = width or get_w(height, clip.width / clip.height)
-
-    clip_y = depth(get_y(clip), 32)
-
-    descale = _perform_descale(Resolution(width, height), clip_y, kernel)
-    rescaled = depth(core.std.PlaneStats(descale.rescaled, clip_y), get_depth(clip))
-
-    if clip.format.num_planes == 1:
-        rescaled = core.text.FrameProps(rescaled, "PlaneStatsDiff") if show_error else rescaled
+    if isinstance(smooth, int):
+        filter_func = partial(core.std.BoxBlur, hradius=smooth, vradius=smooth)
+    elif isinstance(smooth, float):
+        filter_func = partial(core.tcanny.TCanny, sigma=smooth, mode=-1)
     else:
-        merge = core.std.ShufflePlanes([rescaled, clip], [0, 1, 2], vs.YUV)
-        rescaled = core.text.FrameProps(merge, "PlaneStatsDiff") if show_error else merge
+        filter_func = smooth  # type: ignore[assignment]
 
-    return rescaled, descale
+    if width is None:
+        width = get_w(height, aspect_ratio=clip.width/clip.height)
+
+    clip = depth(clip, 32)
+
+    if gamma:
+        clip = gamma2linear(clip, curve, sigmoid=sigmoid, epsilon=epsilon)
+
+    l1 = kernel.scale(clip, width, height)
+    l2 = kernel.scale(clip.std.Expr('x dup *'), width, height)
+
+    m = filter_func(l1)
+
+    sl_plus_m_square = filter_func(l1.std.Expr('x dup *'))
+    sh_plus_m_square = filter_func(l2)
+    m_square = m.std.Expr('x dup *')
+    r = core.std.Expr([sl_plus_m_square, sh_plus_m_square, m_square], f'x z - {epsilon} < 0 y z - x z - / sqrt ?')
+    t = filter_func(core.std.Expr([r, m], 'x y *'))
+    m = filter_func(m)
+    r = filter_func(r)
+    d = core.std.Expr([m, r, l1, t], 'x y z * + a -')
+
+    if gamma:
+        d = linear2gamma(d, curve, sigmoid=sigmoid)
+
+    return d
+
+
+@disallow_variable_format
+@disallow_variable_resolution
+def gamma2linear(clip: vs.VideoNode, curve: CURVES, gcor: float = 1.0,
+                 sigmoid: bool = False, thr: float = 0.5, cont: float = 6.5,
+                 epsilon: float = 1e-6) -> vs.VideoNode:
+    assert clip.format
+
+    if get_depth(clip) != 32 and clip.format.sample_type != vs.FLOAT:
+        raise ValueError('Only 32 bits float is allowed')
+
+    c = get_coefs(curve)
+
+    expr = f'x {c.k0} <= x {c.phi} / x {c.alpha} + 1 {c.alpha} + / {c.gamma} pow ? {gcor} pow'
+    if sigmoid:
+        x0 = f'1 1 {cont} {thr} * exp + /'
+        x1 = f'1 1 {cont} {thr} 1 - * exp + /'
+        expr = f'{thr} 1 {expr} {x1} {x0} - * {x0} + {epsilon} max / 1 - {epsilon} max log {cont} / -'
+
+    expr = f'{expr} 0.0 max 1.0 min'
+
+    return core.std.Expr(clip, expr).std.SetFrameProps(_Transfer=8)
+
+
+@disallow_variable_format
+@disallow_variable_resolution
+def linear2gamma(clip: vs.VideoNode, curve: CURVES, gcor: float = 1.0,
+                 sigmoid: bool = False, thr: float = 0.5, cont: float = 6.5,
+                 ) -> vs.VideoNode:
+    assert clip.format
+
+    if get_depth(clip) != 32 and clip.format.sample_type != vs.FLOAT:
+        raise ValueError('Only 32 bits float is allowed')
+
+    c = get_coefs(curve)
+
+    expr = 'x'
+    if sigmoid:
+        x0 = f'1 1 {cont} {thr} * exp + /'
+        x1 = f'1 1 {cont} {thr} 1 - * exp + /'
+        expr = f'1 1 {cont} {thr} {expr} - * exp + / {x0} - {x1} {x0} - /'
+
+    expr += f' {gcor} pow'
+    expr = f'{expr} {c.k0} {c.phi} / <= {expr} {c.phi} * {expr} 1 {c.gamma} / pow {c.alpha} 1 + * {c.alpha} - ?'
+    expr = f'{expr} 0.0 max 1.0 min'
+
+    return core.std.Expr(clip, expr).std.SetFrameProps(_Transfer=curve)
+
+
+@disallow_variable_format
+@disallow_variable_resolution
+def comparative_descale(clip: vs.VideoNode, width: int | None = None, height: int = 720,
+                        kernel: Kernel | None = None, thr: float = 5e-8) -> vs.VideoNode:
+    """
+    Easy wrapper to descale to SharpBicubic and an additional kernel,
+    compare them, and then pick one or the other.
+
+    The output clip has props that can be used to frameeval specific kernels by the user.
+
+    :param clip:        Input clip
+    :param width:       Width to descale to (if None, auto-calculated)
+    :param height:      Descale height
+    :param kernel:      Kernel to compare BicubicSharp to (Default: Spline36 if None)
+    :param thr:         Threshold for which kernel to pick
+
+    :return:            Descaled clip
+    """
+    def _compare(n: int, f: vs.VideoFrame, sharp: vs.VideoNode, other: vs.VideoNode) -> vs.VideoNode:
+        sharp_diff = get_prop(f[0], 'PlaneStatsDiff', float)  # type:ignore[arg-type]
+        other_diff = get_prop(f[1], 'PlaneStatsDiff', float)  # type:ignore[arg-type]
+
+        return sharp if other_diff - thr > sharp_diff else other
+
+    bsharp = BicubicSharp()
+    kernel = kernel or Spline36()
+
+    if isinstance(kernel, Bicubic) and bsharp.b == kernel.b and bsharp.c == kernel.c:
+        raise ValueError("comparative_descale: 'You may not compare BicubicSharp with itself!'")
+
+    if width is None:
+        width = get_w(height, aspect_ratio=clip.width/clip.height)
+
+    clip_y = get_y(clip)
+    # TODO: Add support for multiple scaler combos. Gotta rethink how `thr` will work tho
+    sharp = bsharp.descale(clip_y, width, height)
+    sharp_up = bsharp.scale(sharp, clip.width, clip.height)
+
+    # TODO: Fix so you can also pass additional params to object. Breaks currently (not callable)
+    other = kernel.descale(clip_y, width, height)
+    other_up = kernel.scale(other, clip.width, clip.height)
+
+    # We need a diff between the rescaled clips and the original clip
+    sharp_diff = core.std.PlaneStats(sharp_up, clip_y)
+    other_diff = core.std.PlaneStats(other_up, clip_y)
+
+    # Extra props for future frame evalling in case it might prove useful (credit masking, for example)
+    sharp = sharp.std.SetFrameProp('scaler', data=bsharp.__class__.__name__)
+    other = other.std.SetFrameProp('scaler', data=kernel.__class__.__name__)
+
+    return core.std.FrameEval(sharp, partial(_compare, sharp=sharp, other=other), [sharp_diff, other_diff])
+
+
+@disallow_variable_format
+@disallow_variable_resolution
+def comparative_restore(clip: vs.VideoNode, width: int | None = None, height: int = 720,
+                        kernel: Kernel | None = None) -> vs.VideoNode:
+    """
+    Companion function to go with comparative_descale
+    to reupscale the clip for descale detail masking.
+
+    :param clip:        Input clip
+    :param width:       Width to upscale to (if None, auto-calculated)
+    :param height:      Upscale height
+    :param kernel:      Kernel to compare BicubicSharp to (Default: Spline36 if None)
+    """
+    bsharp = BicubicSharp()
+    kernel = kernel or Spline36()
+
+    if isinstance(kernel, Bicubic) and bsharp.b == kernel.b and bsharp.c == kernel.c:
+        raise ValueError("comparative_restore: 'You may not compare BicubicSharp with itself!'")
+
+    if width is None:
+        width = get_w(height, aspect_ratio=clip.width/clip.height)
+
+    def _compare(n: int, f: vs.VideoFrame, sharp_up: vs.VideoNode, other_up: vs.VideoNode) -> vs.VideoNode:
+        return sharp_up if get_prop(f, 'scaler', bytes) == b'BicubicSharp' else other_up
+
+    # TODO: just read prop and automatically figure out scaler TBH
+    sharp_up = bsharp.scale(clip, width, height)
+    other_up = kernel.scale(clip, width, height)
+
+    return core.std.FrameEval(sharp_up, partial(_compare, sharp_up=sharp_up, other_up=other_up), clip)
 
 
 # TODO: Write a function that checks every possible combination of B and C in bicubic
